"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Defines custom options used internally at Modal.
Custom options must be in the range 50000-99999.
Reference: https://protobuf.dev/programming-guides/proto2/#customoptions
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.descriptor_pb2
import google.protobuf.internal.extension_dict

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

AUDIT_TARGET_ATTR_FIELD_NUMBER: builtins.int
AUDIT_EVENT_NAME_FIELD_NUMBER: builtins.int
AUDIT_EVENT_DESCRIPTION_FIELD_NUMBER: builtins.int
audit_target_attr: google.protobuf.internal.extension_dict._ExtensionFieldDescriptor[google.protobuf.descriptor_pb2.FieldOptions, builtins.bool]
audit_event_name: google.protobuf.internal.extension_dict._ExtensionFieldDescriptor[google.protobuf.descriptor_pb2.MethodOptions, builtins.str]
audit_event_description: google.protobuf.internal.extension_dict._ExtensionFieldDescriptor[google.protobuf.descriptor_pb2.MethodOptions, builtins.str]
