"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Defines custom options used internally at Modal.
Custom options must be in the range 50000-99999.
Reference: https://protobuf.dev/programming-guides/proto2/#customoptions
"""
